@@ -12,11 +12,12 @@ or 'personal' environment. The defaults are not meant for
 production use.
 
 To customize local client configuration, add/update
-`~/.metrique/http_api.json` (default).
+`~/.metrique/metrique.json` (default).
 
 Paths are UNIX compatible only.
 '''
 
+# FIXME: requires psutil
 import multiprocessing
 import os
 import re
@@ -27,11 +28,13 @@ USER_DIR = os.path.expanduser('~/.metrique')
 CONFIG_DIR = os.path.join(USER_DIR, 'etc')
 LOG_DIR = os.path.join(USER_DIR, 'logs')
 JOURNAL_DIR = os.path.join(USER_DIR, 'journal')
+TMP_DIR = os.path.join(USER_DIR, 'tmp')
+GNUPG_DIR = os.path.expanduser('~/.gnupg')
+HDF5_DIR = os.path.join(USER_DIR, 'hdf5')
 COOKIEJAR = os.path.join(USER_DIR, '.cookiejar')
-
 DEFAULT_CONFIG = os.path.join(CONFIG_DIR, 'metrique')
 
-for path in [USER_DIR, CONFIG_DIR, LOG_DIR, JOURNAL_DIR]:
+for path in [USER_DIR, CONFIG_DIR, LOG_DIR, JOURNAL_DIR, TMP_DIR, HDF5_DIR]:
     if not os.path.exists(path):
         os.makedirs(path)
 
@@ -52,8 +55,9 @@ class Config(JSONConf):
         ssl: Connect with SSL (https)
         ssl_verify: ...
     '''
+    default_config = DEFAULT_CONFIG
+
     def __init__(self, config_file=None, *args, **kwargs):
-        self.default_config = DEFAULT_CONFIG
         self.defaults = {
             'api_version': 'v2',
             'api_rel_path': 'api/v2',
@@ -62,8 +66,13 @@ class Config(JSONConf):
             'batch_size': 250,
             'cookiejar': COOKIEJAR,
             'configdir': CONFIG_DIR,
+            'cube_pkgs': ['cubes'],
+            'cube_paths': [],
             'debug': None,
+            'gnupg_dir': GNUPG_DIR,
+            'gnupg_fingerprint': None,
             'host': '127.0.0.1',
+            'hdf5_dir': HDF5_DIR,
             'journal': True,
             'journaldir': JOURNAL_DIR,
             'logdir': LOG_DIR,
@@ -78,13 +87,35 @@ class Config(JSONConf):
             'sql_batch_size': 250,
             'ssl': False,
             'ssl_verify': True,
+            'tmpdir': TMP_DIR,
             'username': os.getenv('USER'),
             'userdir': USER_DIR,
-            'cube_pkgs': ['cubes'],
-            'cube_paths': [],
         }
         super(Config, self).__init__(config_file=config_file, *args, **kwargs)
 
+    @property
+    def gnupg(self):
+        if hasattr(self, '_gnupg'):
+            gpg = self._gnupg
+        else:
+            # avoid exception in py2.6
+            # workaround until
+            # https://github.com/isislovecruft/python-gnupg/pull/36 is resolved
+            try:
+                from gnupg import GPG
+            except (ImportError, AttributeError):
+                gpg = None
+            else:
+                gpg = GPG(homedir=os.path.expanduser(self['gnupg_dir']))
+        return gpg
+
+    @property
+    def gnupg_pubkey(self):
+        if self.gnupg:
+            return self.gnupg.export_keys(self['gnupg_fingerprint'])
+        else:
+            return ''
+
     @property
     def api_url(self):
         ''' Url and schema - http(s)? needed to call metrique api '''
