@@ -34,13 +34,13 @@ and more.
     valid date format: '%Y-%m-%d %H:%M:%S,%f', '%Y-%m-%d %H:%M:%S', '%Y-%m-%d'
 
 '''
-
-
 from copy import copy
-import logging
-from functools import partial
-import os
 import cPickle
+from functools import partial
+import glob
+import logging
+import os
+import pandas as pd
 import requests
 import simplejson as json
 import urllib
@@ -55,9 +55,293 @@ logging.basicConfig()
 root_logger = logging.getLogger()
 [root_logger.removeHandler(hdlr) for hdlr in root_logger.handlers]
 BASIC_FORMAT = "%(name)s:%(message)s"
+FILETYPES = {'csv': pd.read_csv,
+             'json': pd.read_json}
 
 
-class HTTPClient(object):
+# BaseClient should have a base-extract (default: in-memory)
+# which HTTPClient (etc) should call if the client's config
+# indicates 'class_client': 'HTTP' or 'H5' or 'Base'
+class BaseClient(object):
+    '''
+    Essentially, a cube is a list of dictionaries.
+    '''
+    name = 'NO_NAME'
+    ' defaults is frequently overrided in subclasses as a property '
+    defaults = None
+    ' fields is frequently overrided in subclasses as a property too '
+    fields = None
+    ' filename of the data when saved to disk '
+    saveas = ''
+    ' a place to put stuff, temporarily... '
+    _cache = None
+
+    def __new__(cls, *args, **kwargs):
+        '''
+        Return the specific cube class, if specified. Its
+        expected the cube will be available in sys.path.
+
+        If the cube fails to import, just move on.
+
+            >>> import pyclient
+            >>> c = pyclient(cube='git_commit')
+                <type HTTPClient(...)>
+        '''
+        if 'cube' in kwargs and kwargs['cube']:
+            cls = get_cube(cube=kwargs['cube'], init=False)
+        else:
+            cls = cls
+        return object.__new__(cls)
+
+    def __init__(self, config_file=None, name=None, objects=None, **kwargs):
+        # don't assign to {} in class def, define here to avoid
+        # multiple pyclient objects linking to a shared dict
+        if self.defaults is None:
+            self.defaults = {}
+        if self.fields is None:
+            self.fields = {}
+        if self._cache is None:
+            self._cache = {}
+
+        self._config_file = config_file or Config.default_config
+
+        # all defaults are loaded, unless specified in
+        # metrique_config.json
+        self.load_config()
+
+        # update config object with any additional kwargs
+        for k, v in kwargs.items():
+            if v is not None:
+                self.config[k] = v
+
+        # set name if passed in, but don't overwrite default if not
+        if name is not None:
+            self.name = name
+        if not self.name:
+            raise RuntimeError("cube requires a name")
+
+        if objects:
+            # set objects if passed in, but don't overwrite default if not
+            self.objects = objects
+
+        self.config.logdir = os.path.expanduser(self.config.logdir)
+        if not os.path.exists(self.config.logdir):
+            os.makedirs(self.config.logdir)
+        self.config.logfile = os.path.join(self.config.logdir,
+                                           self.config.logfile)
+
+        # keep logging local to the cube so multiple
+        # cubes can independently log without interferring
+        # with each others logging.
+        self.debug_set()
+
+#################### special cube object handlers #######################
+    def __getitem__(self, name):
+        return self.objects[name]
+
+    def __len__(self):
+        return len(self.objects)
+
+    def __iter__(self):
+        return iter(self.objects)
+
+    def __next__(self):
+        yield next(self)
+
+    def next(self):
+        return self.__next__()
+
+    def __getslice__(self, i, j):
+        return self.objects[i:j]
+
+    def __contains__(self, item):
+        return item in self.objects
+
+    def oids(self):
+        raise NotImplementedError
+
+    def keys(self):
+        return self.objects.keys()
+
+    def items(self):
+        return self.objects.items()
+
+    def __str__(self):
+        return str(self.objects)
+
+    def __repr__(self):
+        return repr(self.objects)
+
+####################### pandas/hd5 python interface ################
+    @property
+    def df(self):
+        if isinstance(self.objects, pd.DataFrame):
+            return self.objects
+        elif isinstance(self.objects, pd.HDFStore):
+            # always expected under the 'objects' hdfs5 group
+            return self.objects['objects']
+        else:
+            return pd.DataFrame(self.objects)
+
+    def load_hdf5(self, path, filetype=None, **kwargs):
+        '''
+        cache to hd5 on disk
+        '''
+        # kwargs are for passing ftype load options (csv.delimiter, etc)
+        # expect the use of globs; eg, file* might result in fileN (file1,
+        # file2, file3), etc
+        datasets = glob.glob(os.path.expanduser(path))
+        for ds in datasets:
+            if os.path.exists(ds):
+                filetype = path.split('.')[-1]
+                # buid up a single dataframe by concatting
+                # all globbed files together
+                self.objects = pd.concat(
+                    [self._load_hdf5(ds, filetype, **kwargs)
+                     for ds in datasets])
+            else:
+                self.objects = pd.DataFrame()
+        return self.objects
+
+    def _load_hdf5(self, path, filetype, **kwargs):
+        if filetype == 'json':
+            objects = pd.read_json(path, **kwargs)
+        elif filetype in ('csv', 'txt'):
+            # load the file into hdf5, according to filetype
+            objects = pd.read_csv(path, **kwargs)
+        elif filetype == 'xls':
+            # load the file into hdf5, according to filetype
+            exfile = pd.ExcelFile(path, **kwargs)
+            objects = dict([(s, exfile.parse(s)) for s in exfile.sheet_names])
+        else:
+            raise TypeError("Invalid filetype: %s" % filetype)
+        return objects
+
+    @property
+    def hdf5(self):
+        store = self._cache.get('hdf5_store', pd.HDFStore(self.hdf5_path))
+        objects = getattr(store, 'objects', pd.DataFrame())
+        build = getattr(self, '_build', True)
+        if objects and build:
+            df = pd.concat((objects, self.df))
+            store['objects'] = df
+            self._build = False
+        else:
+            store['objects'] = self.df
+        self._cache['hdf5_store'] = store
+        return store
+
+    @property
+    def hdf5_path(self):
+        filename = '__'.join((self.owner, self.name)) + '.hd5'
+        return os.path.join(self.config.hdf5_dir, filename)
+
+#################### misc #######################
+    def debug_set(self, level=None, logstdout=None, logfile=None):
+        '''
+        if we get a level of 2, we want to apply the
+        debug level to all loggers
+        '''
+        if level is None:
+            level = self.config.debug
+        if logstdout is None:
+            logstdout = self.config.logstdout
+        if logfile is None:
+            logfile = self.config.logfile
+
+        basic_format = logging.Formatter(BASIC_FORMAT)
+
+        if level == 2:
+            self._logger_name = None
+            logger = logging.getLogger()
+            logger = self._debug_set_level(logger, level)
+
+        self._logger_name = '%s.%s' % ('metrique', self.name)
+        logger = logging.getLogger(self._logger_name)
+        logger.propagate = 0
+
+        logger.handlers = []  # reset handlers
+        if logstdout:
+            hdlr = logging.StreamHandler()
+            hdlr.setFormatter(basic_format)
+            logger.addHandler(hdlr)
+
+        if self.config.log2file and logfile:
+            hdlr = logging.FileHandler(logfile)
+            hdlr.setFormatter(basic_format)
+            logger.addHandler(hdlr)
+
+        logger = self._debug_set_level(logger, level)
+        self.logger = logger
+
+    def _debug_set_level(self, logger, level):
+        if level in [-1, False]:
+            logger.setLevel(logging.WARN)
+        elif level in [0, None]:
+            logger.setLevel(logging.INFO)
+        elif level in [True, 1, 2]:
+            logger.setLevel(logging.DEBUG)
+        return logger
+
+    def get_cube(self, cube, init=True, name=None, **kwargs):
+        ' wrapper for utils.get_cube(); try to load a cube, pyclient '
+        config = copy(self.config)
+        # don't apply the name to the current obj, but to the object
+        # we get back from get_cube
+        return get_cube(cube=cube, init=init, config=config,
+                        name=name, **kwargs)
+
+    def load_config(self, config=None):
+        ' try to load a config file and handle when its not available '
+        if type(config) is type(Config):
+            self._config_file = config.config_file
+        else:
+            config_file = config or self._config_file
+            self.config = Config(config_file=config_file)
+            self._config_file = config_file
+
+    #############################################################
+    # FIXME: objects should be a HDFStore! not an in memory dict
+    @property
+    def objects(self):
+        ''' always return a list of dicts '''
+        objs = self._cache.get('objects', {})
+        if not objs:
+            return {}
+        if isinstance(objs, pd.HDFStore):
+            objs = objs['objects']
+        if isinstance(objs, pd.DataFrame):
+            df = objs
+            objs = df.T.to_dict().values()
+        return objs
+
+    @objects.setter
+    def objects(self, value):
+        if isinstance(value, pd.HDFStore):
+            value = value['objects']
+        if isinstance(value, pd.DataFrame):
+            df = value
+        else:
+            df = pd.DataFrame(value)
+        objects = df.T.to_dict().values()
+        self._cache['objects'] = objects
+        # dependencies of objects should refresh
+        self._build = True
+        return objects
+
+    @property
+    def objectsi(self):
+        return iter(self.objects)
+
+    def save_uri(self, uri, saveas):
+        return urllib.urlretrieve(uri, saveas)
+
+    def whoami(self, auth=False):
+        ' quick way of checking the username the instance is working as '
+        return self.config['username']
+
+
+class HTTPClient(BaseClient):
     '''
     This is the main client bindings for metrique http
     rest api.
@@ -67,12 +351,6 @@ class HTTPClient(object):
 
 
     '''
-    name = None
-    ' defaults is frequently overrided in subclasses as a property '
-    defaults = {}
-    ' fields is frequently overrided in subclasses as a property too '
-    fields = {}
-
     # frequently 'typed' commands have shorter aliases too
     user_aboutme = aboutme = user_api.aboutme
     user_login = login = user_api.login
@@ -107,164 +385,14 @@ class HTTPClient(object):
     query_sample = sample = query_api.sample
     query_aggregate = aggregate = query_api.aggregate
 
-    def __new__(cls, *args, **kwargs):
-        '''
-        Return the specific cube class, if specified. Its
-        expected the cube will be available in sys.path.
-
-        If the cube fails to import, just move on.
-
-            >>> import pyclient
-            >>> c = pyclient(cube='git_commit')
-                <type HTTPClient(...)>
-        '''
-        if 'cube' in kwargs and kwargs['cube']:
-            cls = get_cube(cube=kwargs['cube'], init=False)
-        else:
-            cls = cls
-        return object.__new__(cls)
-
-    def __init__(self, cube=None, config_file=None, owner=None,
-                 name=None, **kwargs):
-        self._config_file = config_file
-        # all defaults are loaded, unless specified in
-        # metrique_config.json
-        self.load_config()
-
-        # update config object with any additional kwargs
-        for k, v in kwargs.items():
-            if v is not None:
-                self.config[k] = v
-
+    def __init__(self, cube=None, owner=None, **kwargs):
+        super(HTTPClient, self).__init__(**kwargs)
         self.owner = owner or self.config.username
-
-        if name:
-            # override the cube name
-            self.name = name
-
-        self.config.logdir = os.path.expanduser(self.config.logdir)
-        if not os.path.exists(self.config.logdir):
-            os.makedirs(self.config.logdir)
-        self.config.logfile = os.path.join(self.config.logdir,
-                                           self.config.logfile)
-
-        # keep logging local to the cube so multiple
-        # cubes can independently log without interferring
-        # with each others logging.
-        self.debug_set()
-
         # load a new requests session; for the cookies.
         self._load_session()
         self._auto_login_attempted = False
-        self._cache = {}
-
-    def __getitem__(self, name):
-        if isinstance(name, slice):
-            op = 'in'
-            one = False
-        else:
-            op = '=='
-            one = True
-
-        name_json = json.dumps(name, default=json_encode, ensure_ascii=False)
-        result = self.find('_oid %s %s' % (op, name_json),
-                           fields='__all__',
-                           raw=True, one=one)
-        if not result:
-            if op == 'in':
-                result = []
-            else:
-                result = {}
-        return result
-
-    def __setitem__(self, name, value):
-        raise NotImplementedError
-
-    def __delitem__(self, name):
-        raise NotImplementedError
-
-    def __len__(self):
-        return self.count()
-
-    def __iter__(self):
-        return self
-
-    def insert(self, objs):
-        self.save(objs)
-
-    def extend(self, objs):
-        self.save(objs)
-
-    def next(self):
-        count = self._cache.get('counter', None)
-        if count is None:
-            k = self._cache['count'] = self.count()
-            count = self._cache['counter'] = 0
-            self._cache['sort'] = self.config.sort
-        else:
-            k = self._cache['count']
-        if count < k:
-            docs = self.find(sort=[('_oid', self._cache['sort'])],
-                             skip=count,
-                             limit=self.config.batch_size,
-                             raw=True,
-                             fields='__all__')
-            self._cache['counter'] += len(docs)
-            # FIXME: return back a generator rather than tuple?
-            return docs
-        else:
-            del self._cache['counter']
-            del self._cache['count']
-            del self._cache['sort']
-            raise StopIteration
-
-    def __getslice__(self, i, j):
-        return self.find(sort=[('_oid', self.config.sort)],
-                         raw=True, fields='__all__',
-                         skip=i, limit=j)
-
-    def __contains__(self, item):
-        return bool(self.count('_oid == %s' % item))
-
-    def keys(self):
-        i = 0
-        size = self.config.batch_size
-        for j in xrange(size, self.count(), size):
-            objs = self[i:j]
-            for o in objs:
-                yield o['_oid']
-                i = j
-        else:
-            objs = self[i:self.count()]
-            for o in objs:
-                yield o['_oid']
-
-    def values(self):
-        i = 0
-        size = self.config.batch_size
-        for j in xrange(size, self.count(), size):
-            objs = self[i:j]
-            for o in objs:
-                yield o
-                i = j
-        else:
-            objs = self[i:self.count()]
-            for o in objs:
-                yield o
-
-    def items(self):
-        i = 0
-        size = self.config.batch_size
-        for j in xrange(size, self.count(), size):
-            objs = self[i:j]
-            for o in objs:
-                yield o['_oid'], o
-                i = j
-        else:
-            objs = self[i:self.count()]
-            for o in objs:
-                yield o['_oid'], o
 
+#################### HTTP specific API methods/overrides ###############
     def activity_get(self, ids=None):
         '''
         Returns a dictionary of `id: [(when, field, removed, added)]` kv pairs
@@ -318,58 +446,6 @@ class HTTPClient(object):
         with open(path, 'w') as f:
             cPickle.dump(dfc(self.session.cookies), f)
 
-    @property
-    def current_user(self):
-        ' alias for whoami(); returns back username in config.username '
-        return self.whoami()
-
-    def debug_set(self, level=None, logstdout=None, logfile=None):
-        '''
-        if we get a level of 2, we want to apply the
-        debug level to all loggers
-        '''
-        if level is None:
-            level = self.config.debug
-        if logstdout is None:
-            logstdout = self.config.logstdout
-        if logfile is None:
-            logfile = self.config.logfile
-
-        basic_format = logging.Formatter(BASIC_FORMAT)
-
-        if level == 2:
-            self._logger_name = None
-            logger = logging.getLogger()
-        elif not self.name:
-            self._logger_name = __name__
-            logger = logging.getLogger(self._logger_name)
-            logger.propagate = 0
-        else:
-            self._logger_name = '%s.%s' % (__name__, self.name)
-            logger = logging.getLogger(self._logger_name)
-            logger.propagate = 0
-
-        # reset handlers
-        logger.handlers = []
-
-        if logstdout:
-            hdlr = logging.StreamHandler()
-            hdlr.setFormatter(basic_format)
-            logger.addHandler(hdlr)
-
-        if self.config.log2file and logfile:
-            hdlr = logging.FileHandler(logfile)
-            hdlr.setFormatter(basic_format)
-            logger.addHandler(hdlr)
-
-        if level in [-1, False]:
-            logger.setLevel(logging.WARN)
-        elif level in [0, None]:
-            logger.setLevel(logging.INFO)
-        elif level in [True, 1, 2]:
-            logger.setLevel(logging.DEBUG)
-        self.logger = logger
-
     def _delete(self, *args, **kwargs):
         ' requests DELETE; using current session '
         return self._run(self.session.delete, *args, **kwargs)
@@ -396,17 +472,6 @@ class HTTPClient(object):
         else:
             return os.path.join(owner, cube)
 
-    def get_cube(self, cube, init=True, **kwargs):
-        ' wrapper for utils.get_cube(); try to load a cube, pyclient '
-        config = copy(self.config)
-        # don't apply the name to the current obj, but to the object
-        # we get back from get_cube
-        name = kwargs.get('name')
-        if name:
-            del kwargs['name']
-        return get_cube(cube=cube, init=init, config=config,
-                        name=name, **kwargs)
-
     def get_last_field(self, field):
         '''
         shortcut for querying to get the last field value for
@@ -445,9 +510,12 @@ class HTTPClient(object):
                       allow_redirects=True, stream=False):
         ' wrapper for running a metrique api request; get/post/etc '
         _auto = self.config.auto_login
+        # avoids bug in requests-2.0.1 - pass a dict no RequestsCookieJar
+        # eg, see: https://github.com/kennethreitz/requests/issues/1744
+        dfc = requests.utils.dict_from_cookiejar
         try:
             _response = runner(_url, auth=(username, password),
-                               cookies=self.session.cookies,
+                               cookies=dfc(self.session.cookies),
                                verify=self.config.ssl_verify,
                                allow_redirects=allow_redirects,
                                stream=stream)
@@ -478,15 +546,6 @@ class HTTPClient(object):
                       json.dumps(v, default=json_encode, ensure_ascii=False))
                     for k, v in kwargs.items()])
 
-    def load_config(self, config=None):
-        ' try to load a config file and handle when its not available '
-        if type(config) is type(Config):
-            self._config_file = config.config_file
-        else:
-            config_file = config or self._config_file
-            self.config = Config(config_file=config_file)
-            self._config_file = config_file
-
     def _load_session(self):
         ' load a fresh new requests session; mainly, reset cookies '
         self.session = requests.Session()
@@ -541,9 +600,6 @@ class HTTPClient(object):
                 self.logger.error(content)
                 raise
 
-    def save_uri(self, uri, saveas):
-        return urllib.urlretrieve(uri, saveas)
-
     def _save(self, filename, *args, **kwargs):
         ' requests GET of a "file stream" using current session '
         return self._run(self.session.get, stream=True, filename=filename,
@@ -554,7 +610,7 @@ class HTTPClient(object):
         if auth:
             self.user_login()
         else:
-            return self.config['username']
+            super(HTTPClient, self).whoami()
 
 
 # import alias
